﻿--[[ Lua Hunspell ]]--

----------------------------------------
--[[ description:
  -- Spell checking based on Hunspell.
  -- Проверка орфографии на основе Hunspell.
--]]
----------------------------------------
--[[ uses:
  LuaFAR,
  far2.
  -- areas: editor.
--]]
----------------------------------------
--[[ based on:
  spell.lua
  (Макрос для проверки орфографии в редакторе.)
  (Searchable Menu.)
  (c) 2013+, zg.
  URL: http://forum.farmanager.com/viewtopic.php?f=60&t=8089&hilit=LuaSpell
--]]
----------------------------------------
--[[ required:
  1. Библиотека dll из проекта http://nhunspell.sourceforge.net/
     URL: http://sourceforge.net/projects/nhunspell/files/
     Отсюда берётся архив с последней версией.
     Из него нужны либо Hunspellx86.dll, либо Hunspellx64.dll.
     Именно эти dll чисто нативные, никакого .NET в них нет.
     Положить их куда-нибудь, откуда они загрузятся, например, в папку с FAR.
  2. Словари.
     По умолчанию макрос настроен на следующие словари:
     ru_RU_yo, en_US, el_GR (нужны файлы aff и dic).
     По умолчанию словари ожидаются в ~%FARPROFILE%\Dictionaries~.
--]]
----------------------------------------
--[[ History:
2014-05-03, v.3.0.1.1
―――――――
[*] Добавлены описания к используемым событиям (спасибо Shmuel'у).

2014-05-02, v.3.0.1.0
―――――――
[*] По умолчанию одновременно используется три словаря:
    русский (с буквой ё), английский и греческий.
[*] При проверке игнорируется комбинированный символ ударения.
    Дополнительные символы и условия можно прописать в ClearWord.
[-] Однобуквенные слова не проверяются (см. строки c WARN).
[-] Ограничен набор букв в составе проверяемых слов.
    Поменять набор можно, изменив локальную переменную CharSet.
[*] Изменены комбинации клавиш: CtrlF12 и CtrlShiftF12.
    Не забудьте поменять на свои используемые комбинации.
--]]
--------------------------------------------------------------------------------

----------------------------------------
local F = far.Flags
local regex = regex

---------------------------------------- config
local DicPath = [[%FARPROFILE%\Dictionaries\]]
local Flag4BIT = bit64.bor(F.FCF_FG_4BIT, F.FCF_BG_4BIT)

local config = {
  -- rus
  {
    dic = DicPath.."ru_RU_yo.dic",
    aff = DicPath.."ru_RU_yo.aff",
    find = [[/[А-Яа-яЁё]+/]],
    color = {
      Flags = Flag4BIT,
      ForegroundColor = 0xF,
      BackgroundColor = 0x4,
    },
    active = true,
  },
  -- eng
  {
    dic = DicPath.."en_US.dic",
    aff = DicPath.."en_US.aff",
    find = [[/[A-Za-z]+/]],
    color = {
      Flags = Flag4BIT,
      ForegroundColor = 0xF,
      BackgroundColor = 0x5,
    },
    active = true,
  },
  -- grk
  {
    dic = DicPath.."el_GR.dic",
    aff = DicPath.."el_GR.aff",
    find = [[/[΄-Ͽἀ-῿]+/]],
    color = {
      Flags = Flag4BIT,
      ForegroundColor = 0xF,
      BackgroundColor = 0x6,
    },
    active = true,
  },
} ---

----------------------------------------
local editors = {}
local colorguid = win.Uuid("46CCE102-965A-4e2d-8263-B59F268C74C8")
local menu_guid = win.Uuid("ECD10910-8CC6-4685-AA8D-7D7413DD7D06")
local active = false

---------------------------------------- Hunspell through FFI
local ffi = require 'ffi'
local C = ffi.C
ffi.cdef[[
void* HunspellInit(void* affixBuffer,size_t affixBufferSize,void* dictionaryBuffer,size_t dictionaryBufferSize,wchar_t* key);
void HunspellFree(void* handle);
bool HunspellSpell(void* handle,const wchar_t* word);
wchar_t** HunspellSuggest(void* handle,const wchar_t* word);
int lstrlenW(wchar_t* lpString);
]]
local hunspell = ffi.load("hunspell"..(win.GetEnv("PROCESSOR_ARCHITECTURE"):gsub("AMD64","x64")))

---------------------------------------- Init

local function LoadFile (filename)
  local file = io.open(filename, "rb")
  if not file then return end

  local data = file:read("*a")
  file:close()

  local result = ffi.new("uint8_t[?]", #data)
  ffi.copy(result, data, #data)

  return result, #data
end -- LoadFile

local function ExpandEnv (s)
  return s:gsub("%%(.-)%%", win.GetEnv)
end

local function Init ()

  for _, v in ipairs(config) do
    v.dic_data, v.dic_len = LoadFile(ExpandEnv(v.dic))
    v.aff_data, v.aff_len = LoadFile(ExpandEnv(v.aff))
    if v.dic_data and v.aff_data then
      v.handle = hunspell.HunspellInit(v.aff_data, v.aff_len, v.dic_data, v.dic_len, nil)
      v.regex = regex.new(v.find)
    else
      v.active = false
    end
  end

  Init = function() end
end -- Init

---------------------------------------- Menu

local function ShowMenu (strings, wordLen)
  local info = editor.GetInfo()
  local menuShadowWidth = 2
  local menuOverheadWidth = menuShadowWidth + 4 --6 => 2 рамка, 2 тень, 2 место для чекмарка
  local menuOverheadHeight = 3 --3 => 2 рамка, 1 тень
  local menuWidth = 0

  local Items = {}
  for _, v in ipairs(strings) do
    menuWidth = math.max(menuWidth, v:len())
    table.insert(Items, {Flags = 0, Text = v })
  end

  local menuHeight = 1
  local coorX = info.CurTabPos - info.LeftPos
  local coorY = info.CurLine - info.TopScreenLine
  local menuX = math.max(0, coorX - menuWidth - menuOverheadWidth + menuShadowWidth)
  menuX = (info.WindowSizeX - coorX) > (coorX + 2 - wordLen) and (coorX + 1) or menuX -- меню справа или слева от слова?
  local menuY = 0
  if (info.WindowSizeY - coorY - 1) > (coorY + 1) then -- меню сверху или снизу?
    -- снизу
    menuY = coorY + 2
    menuHeight = info.WindowSizeY - menuY + 1 - menuOverheadHeight
    menuHeight = math.min(menuHeight, #strings)
  else
    -- сверху
    menuY = coorY - #strings - 1
    if menuY < 1 then menuY = 1 end
    menuHeight = coorY - menuY - 1
  end

  -- fix menu width
  if (menuX + menuWidth + menuOverheadWidth) > info.WindowSizeX then
    menuWidth = info.WindowSizeX - menuX - menuOverheadWidth
  end

  local Form = {
    {"DI_LISTBOX", 0, 0, menuWidth + 3, menuHeight + 1, Items, 0, 0, 0, ""}
  }

  local function DlgProc (dlg, msg, param1, param2)
  end -- DlgProc

  local hDlg = far.DialogInit(menu_guid,
                              menuX, menuY,
                              menuX + menuWidth + 3,
                              menuY + menuHeight + 1,
                              nil, Form, 0, DlgProc)
  local Index = far.DialogRun(hDlg) > 0 and
                far.SendDlgMessage(hDlg, F.DM_LISTGETCURPOS, 1).SelectPos or nil
  far.DialogFree(hDlg)

  return Index
end -- ShowMenu

---------------------------------------- Spell
local CharSet = "[A-Za-zÀ-ÖØ-ßà-öøÿĀ-ʯḀ-ỿА-Яа-яЁёЎўЄєЇї΄-Ͽἀ-῿_0-9́]+"
local InnerCharSet = CharSet.."$"
local StartCharSet = "^"..CharSet

-- Преобразование строки в WChar.
local function StringToWChar (s)
  local s = win.Utf8ToUtf16(s)
  local result = ffi.new("wchar_t[?]", #s / 2 + 1)
  ffi.copy(result, s)
  return result
end -- StringToWChar

-- Преобразование WChar в строку.
local function WCharToString (s)
  return win.Utf16ToUtf8( ffi.string(s, 2 * C.lstrlenW(s)) )
end -- WCharToString

local function ClearWord (word)
  return word:gsub("́", "")
end -- ClearWord

local function CheckSpell ()
  Init()

  local Info = editor.GetInfo(-1)
  if not Info then return end

  local row, pos = Info.CurLine, Info.CurPos
  local line, eol = editor.GetString(Info.EditorID, row, 3)
  if not line then return end

  local word = ""
  local s, spos = line, 0
  if pos <= s:len() + 1 then
    local slab = pos > 1 and s:sub(1, pos - 1):match(InnerCharSet) or ""
    local tail = s:sub(pos):match(StartCharSet) or ""
    spos = pos - slab:len()
    word = slab..tail
  end
  local wLen = word:len()
  word = ClearWord(word)
  if word:len() <= 1 then return end -- WARN: exclude one char words

  local wData = StringToWChar(word)
  for _, v in ipairs(config) do
    if v.active and v.regex:match(word) and
       not hunspell.HunspellSpell(v.handle, wData) then

      local suggestions = hunspell.HunspellSuggest(v.handle, wData)
      local k, items = 0, {}
      while suggestions[k] ~= ffi.NULL do
        table.insert(items, WCharToString(suggestions[k]))
        k = k + 1
      end

      if #items > 0 then
        local Index = ShowMenu(items, wLen)
        if Index then
          local s = line:sub(1, spos - 1)..
                    items[Index]..
                    line:sub(spos + wLen)
          editor.SetString(-1, 0, s, eol)
        end
      end

      break
    end
  end

end -- CheckSpell

---------------------------------------- Remove

function GetData (id)
  local data = editors[id]
  if not data then
    editors[id] = { start = 0, finish = -1, }
    data = editors[id]
  end

  return data
end ---- GetData

local function RemoveColors (id)
  local data = GetData(id)
  for k = data.start, data.finish do
    editor.DelColor(id, k, 0, colorguid)
  end

  data.start = 0
  data.finish = -1
  return data
end -- RemoveColors

local Far_WinInfo = F.ACTL_GETWINDOWINFO
local WinType_Editor = F.WTYPE_EDITOR

local function RemoveAllColors ()
  local Count = far.AdvControl(F.ACTL_GETWINDOWCOUNT, 0, 0)
  for k = 1, Count do
    local Info = far.AdvControl(Far_WinInfo, k, 0)
    if Info and Info.Type == WinType_Editor then
      RemoveColors(Info.Id)
    end
  end
end -- RemoveAllColors

---------------------------------------- Spell all

local CheckCharSet = "/\\b"..CharSet.."\\b/"

local CheckExcludes = {
} --- CheckExcludes

local Mark_Current = F.ECF_TABMARKCURRENT

local function CheckSpellAll (Info)
  if not Info then return end

  Init()

  local data = RemoveColors(Info.EditorID)
  data.start = Info.TopScreenLine
  data.finish = math.min(Info.TopScreenLine + Info.WindowSizeY - 1, Info.TotalLines)
  local Regex = regex.new(CheckCharSet)

  for k = data.start, data.finish do
    local line = editor.GetString(-1, k, 3)
    local pos = 1
    while true do
      local sbegin, send = Regex:find(line, pos)
      if not sbegin then break end

      pos = send + 1 -- for next word
      if send > sbegin then -- WARN: exclude one char words
        local word = line:sub(sbegin, send)
        word = ClearWord(word)
        if word == "" then return end

        local wData = StringToWChar(word)
        for _, v in ipairs(config) do
          if v.active and v.regex:match(word) and
             not CheckExcludes[word] and
             not hunspell.HunspellSpell(v.handle, wData) then
            editor.AddColor(Info.EditorID, k, sbegin, send,
                            Mark_Current, v.color, 199, colorguid)
            break
          end
        end
      end
    end -- while
  end -- for

end -- CheckSpellAll

---------------------------------------- Events
Event {
  group = "EditorEvent",
  description = "Check spell all",

  action = function (id, event, param)
    if event == F.EE_READ then
      editors[id] = { start = 0, finish = -1, }
    end
    if event == F.EE_CLOSE then
      editors[id] = nil
    end
    if event == F.EE_REDRAW then
      if active then
        CheckSpellAll(editor.GetInfo())
      end
    end
  end,
} -- Event "EditorEvent"

Event {
  group = "ExitFAR",
  description = "Remove all spell colors",

  action = function ()
    RemoveAllColors()

    for _, v in ipairs(config) do
      if v.active and v.handle then
        hunspell.HunspellFree(v.handle)
      end
    end
  end,
} -- Event "ExitFAR"

---------------------------------------- Macros
Macro {
  area = "Editor",
  key = "CtrlF12",
  description = "Check spell",

  action = CheckSpell,
} ---

Macro {
  area = "Editor",
  key = "CtrlShiftF12",
  description = "Highlight spell errors",

  action = function()
    active = not active
    if not active then
      RemoveAllColors()
    end
    editor.Redraw(-1)
  end,
} ---
--------------------------------------------------------------------------------
